//@author: m4d
//@help: a straight port of Francois Tarliers nice ft Cubic Lens Distortion PixelBender shader (also included are some PS 2.0 techniques which feature different subsets of the full PS 3.0 implementation)
//@tags: dispersion, chromatic, abberation, post effect
//@credits: original shader by Francois Tarlier - original source: http://www.francois-tarlier.com/blog/update-cubic-lens-distortion-pixel-bender-shader-for-ae-with-scale-chroamtic-aberration/
//
//    Copyright (c) 2010 Franois Tarlier
//    Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated
//    documentation files (the "Software"), to deal in the Software without restriction, including without limitation
//    the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and
//    to permit persons to whom the Software is furnished to do so, subject to the following conditions:
//
//    The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.


// --------------------------------------------------------------------------------------------------
// PARAMETERS:
// --------------------------------------------------------------------------------------------------

//transforms
float4x4 tW: WORLD;        //the models world matrix
float4x4 tV: VIEW;         //view matrix as set via Renderer (EX9)
float4x4 tP: PROJECTION;
float4x4 tWVP: WORLDVIEWPROJECTION;

//TexCdture
texture Tex <string uiname="Texture";>;
sampler Samp = sampler_state    //sampler for doing the texture-lookup
{
    Texture   = (Tex);          //apply a texture to the sampler
    MipFilter = LINEAR;         //sampler states
    MinFilter = LINEAR;
    MagFilter = LINEAR;
    AddressU = BORDER;
    AddressV = BORDER;
};

//texture transformation marked with semantic TEXTUREMATRIX to achieve symmetric transformations
float4x4 tTex: TEXTUREMATRIX <string uiname="Texture Transform";>;

float  scale <string uiname="Scale";> = (1.0);
float  cubicDistortion <string uiname="Cubic Distortion"; float uimin=0.0; float uimax=1.0;> = (0.0);
float  distortion <string uiname="Distortion"; float uimin=0.0; float uimax=1.0;> = (0.0);
float  chromaticDispersionRC <string uiname="Chromatic Dispersion Red/Cyan"; float uimin=0.0; float uimax=1.0;> = (0.0);
float  chromaticDispersionBY <string uiname="Chromatic Dispersion Blue/Yellow"; float uimin=0.0; float uimax=1.0;> = (0.0);
bool   invertDistort <string uiname="Invert Distortion"; float uimin=0.0; float uimax=1.0;> = (0.0);
float2 maxSize <string uiname="ScreenSize";> = (1.0,1.0);

//the data structure: "vertexshader to pixelshader"
//used as output data with the VS function
//and as input data with the PS function
struct vs2ps
{
    float4 Pos  : POSITION;
    float2 TexCd : TEXCOORD0;
};

// --------------------------------------------------------------------------------------------------
// VERTEXSHADERS
// --------------------------------------------------------------------------------------------------
vs2ps VS(
    float4 PosO  : POSITION,
    float4 TexCd : TEXCOORD0)
{
    //declare output struct
    vs2ps Out;

    //transform position
    Out.Pos = mul(PosO, tWVP);
    
    //transform texturecoordinates
    Out.TexCd = mul(TexCd, tTex);

    return Out;
}

// --------------------------------------------------------------------------------------------------
// PIXELSHADERS:
// --------------------------------------------------------------------------------------------------

// Compute invert distortion with iterative approach
// (slower but more accurate rather than just inverting k & kcube)
float removeDistortion(float2 tex, float kcube, float f){
float r, rp, rlim, rplim, raw, err, slope;
bool bypass = false;
float ratio = maxSize.x/maxSize.y;
rp = sqrt(ratio * ratio * tex.x * tex.x + tex.y * tex.y);
if (kcube < 0.0) {
    rlim = sqrt((-1.0/3.0) / kcube);
    rplim = rlim * (1.0 + kcube*rlim*rlim);
    if (rp >= 0.99 * rplim) {
        f = rlim / rp;
        bypass = true;
    }
}

if(bypass == false){
    r = rp;
    for (int i = 0; i < 20; i++) {
        raw = kcube * r * r;
        slope = 1.0 + 3.0*raw;
        err = rp - r * (1.0 + raw);
        if (abs(err) < 1.0e-10) {
            i=20;
            }
        r += (err / slope);
        }
    f = r / rp;
    }

    return f;
}


float4 PS3_full(vs2ps In): COLOR
{
    float ratio = maxSize.x/maxSize.y;
    float scaleFactor = 1.0/scale;

    //index of refraction of each color channel, causing chromatic dispersion
    float3 eta = float3(1.0+chromaticDispersionRC*0.09,
                        1.0+chromaticDispersionRC*0.06,
                        1.0+chromaticDispersionRC*0.03);
    float3 eta2 = float3(1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.03);

    float4 inputDistort;
    float2 tex = In.TexCd/0.5 -1;
    
    // lens distortion coefficient
    float k = distortion;
    // cubic distortion value
    float kcube = cubicDistortion;
    float f = 0.0;

    if (invertDistort){
        //REMOVE DISTORTION
        f = removeDistortion(tex,kcube,f);

    }else{
        //APPLY DISTORTION      //*1.7 looks sooo wrong
        float r2 = ratio * ratio * (tex.x) * (tex.x) + (tex.y) * (tex.y);
        //only compute the cubic distortion if necessary
        if( kcube == 0.0){
                f = 1.0 + r2 * k;
        }else{
                f = 1.0 + r2 * (k + kcube * sqrt(r2));
        };
    }

    //Apply CA if necessary
    if (chromaticDispersionRC == 0.0 && chromaticDispersionBY == 0.0){
        float x = f*scaleFactor*(tex.x*0.5)+0.5;
        float y = f*scaleFactor*(tex.y*0.5)+0.5;
        inputDistort = tex2D(Samp,float2(x,y));
    }else{
        float2 rCoords = (f*eta.r*eta2.r)*scaleFactor*(tex.xy*0.5)+0.5;
        float2 gCoords = (f*eta.g*eta2.g)*scaleFactor*(tex.xy*0.5)+0.5;
        float2 bCoords = (f*eta.b*eta2.b)*scaleFactor*(tex.xy*0.5)+0.5;
        float x = f*scaleFactor*(In.TexCd.x*0.5)+0.5;
        float y = f*scaleFactor*(In.TexCd.y*0.5)+0.5;
        inputDistort = float4(tex2D(Samp,rCoords).r , tex2D(Samp,gCoords).g ,
                              tex2D(Samp,bCoords).b, tex2D(Samp,float2(x,y)).a);
    }
    return inputDistort;
}

float4 PS2_noInvert(vs2ps In): COLOR
{
    float ratio = maxSize.x/maxSize.y;
    float scaleFactor = 1.0/scale;

    //index of refraction of each color channel, causing chromatic dispersion
    float3 eta = float3(1.0+chromaticDispersionRC*0.09,
                        1.0+chromaticDispersionRC*0.06,
                        1.0+chromaticDispersionRC*0.03);
    float3 eta2 = float3(1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.03);

    float4 inputDistort;
    float2 tex = In.TexCd.xy;
    tex.x = 2.0*(tex.x/1 - 0.5);
    tex.y = 2.0*(tex.y/1 - 0.5);
    // lens distortion coefficient
    float k = distortion;
    // cubic distortion value
    float kcube = cubicDistortion;
    float f = 0.0;

    //APPLY DISTORTION
    float r2 = ratio * ratio * (tex.x) * (tex.x) + (tex.y) * (tex.y);
    //compute the cubic distortion
    f = 1.0 + r2 * (k + kcube * sqrt(r2));

    //Apply CA
    float2 rCoords = (f*eta.r*eta2.r)*scaleFactor*(tex.xy*0.5)+0.5;
    float2 gCoords = (f*eta.g*eta2.g)*scaleFactor*(tex.xy*0.5)+0.5;
    float2 bCoords = (f*eta.b*eta2.b)*scaleFactor*(tex.xy*0.5)+0.5;
    float x = f*scaleFactor*(In.TexCd.x*0.5)+0.5;
    float y = f*scaleFactor*(In.TexCd.y*0.5)+0.5;

    return float4(tex2D(Samp,rCoords).r, tex2D(Samp,gCoords).g,
                  tex2D(Samp,bCoords).b, tex2D(Samp,float2(x,y)).a);
}


float4 PS2_distortion(vs2ps In): COLOR
{
    float ratio = maxSize.x/maxSize.y;
    float scaleFactor = 1.0/scale;

    //index of refraction of each color channel, causing chromatic dispersion
    float3 eta = float3(1.0+chromaticDispersionRC*0.09,
                        1.0+chromaticDispersionRC*0.06,
                        1.0+chromaticDispersionRC*0.03);
    float3 eta2 = float3(1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.06,
                        1.0+chromaticDispersionBY*0.03);

    float4 inputDistort;
    float2 tex = In.TexCd.xy;
    tex.x = 2.0*(tex.x/1 - 0.5);
    tex.y = 2.0*(tex.y/1 - 0.5);
    // lens distortion coefficient
    float k = distortion;
    // cubic distortion value
    float kcube = cubicDistortion;
    float f = 0.0;

    //APPLY DISTORTION
    float r2 = ratio * ratio * (tex.x) * (tex.x) + (tex.y) * (tex.y);
    //compute the cubic distortion
    f = 1.0 + r2 * (k + kcube * sqrt(r2));


    return float4(tex2D(Samp, (f * eta * eta2)*scaleFactor*(tex.xy*0.5)+0.5));
}

float4 PS2_abberation(vs2ps In): COLOR
{
    //Apply CA
    return float4(tex2D(Samp,In.TexCd+chromaticDispersionRC*0.015).r,
                  tex2D(Samp,In.TexCd+chromaticDispersionRC*0.0075).g,
                  tex2D(Samp,In.TexCd+chromaticDispersionRC*0.00325).b,
                  tex2D(Samp,In.TexCd).a);
}

// --------------------------------------------------------------------------------------------------
// TECHNIQUES:
// --------------------------------------------------------------------------------------------------

technique ft_CubicLensDistortion_ps30
{
    pass P0
    {
        //Wrap0 = U;  // useful when mesh is round like a sphere
        VertexShader = compile vs_1_1 VS();
        PixelShader  = compile ps_3_0 PS3_full();
    }
}

technique ft_CubicLensDistortion_ps20_noInvert
{
    pass P0
    {
        //Wrap0 = U;  // useful when mesh is round like a sphere
        VertexShader = compile vs_1_1 VS();
        PixelShader  = compile ps_2_0 PS2_noInvert();
    }
}

technique ft_CubicLensDistortion_ps20_onlyDistortion
{
    pass P0
    {
        //Wrap0 = U;  // useful when mesh is round like a sphere
        VertexShader = compile vs_1_1 VS();
        PixelShader  = compile ps_2_0 PS2_distortion();
    }
}

technique ft_CubicLensDistortion_ps20_onlyCYAbberation
{
    pass P0
    {
        //Wrap0 = U;  // useful when mesh is round like a sphere
        VertexShader = compile vs_1_1 VS();
        PixelShader  = compile ps_2_0 PS2_abberation();
    }
}

technique TFixedFunction
{
    pass P0
    {
        //transforms
        WorldTransform[0]   = (tW);
        ViewTransform       = (tV);
        ProjectionTransform = (tP);

        //texturing
        Sampler[0] = (Samp);
        TextureTransform[0] = (tTex);
        TexCoordIndex[0] = 0;
        TextureTransformFlags[0] = COUNT2;
        //Wrap0 = U;  // useful when mesh is round like a sphere
        
        Lighting       = FALSE;

        //shaders
        VertexShader = NULL;
        PixelShader  = NULL;
    }
}
